import logging
from logging import NullHandler
from logging.config import dictConfig

from .base.config import LOG
from .core.drivers import ChromeDriver

dictConfig(LOG)
logging.root.setLevel(logging.INFO)
# Set default logging handler to avoid \"No handler found\" warnings.
logging.getLogger(__name__).addHandler(NullHandler())

__all__ = [
    "ChromeDriver",
]



from typing import Union, List, Dict, TypedDict, Any

from weberist.generic.shortcuts import (
    Firefox,
    Chrome,
    Safari,
    Edge,
    SeleniumWebDriver,
    FirefoxOptions,
    ChromeOptions,
    SafariOptions,
    EdgeOptions,
    FirefoxService,
    ChromeService,
    SafariService,
    EdgeService,
    WebElement,
    ChromeDriverManager,
    GeckoDriverManager,
    EdgeChromiumDriverManager
)

WebDriver = Union[Firefox, Chrome, Safari, Edge, SeleniumWebDriver]
WebDriverOptions = Union[
    FirefoxOptions,
    ChromeOptions,
    SafariOptions,
    EdgeOptions,
]
WebDriverServices = Union[
    FirefoxService,
    ChromeService,
    SafariService,
    EdgeService,
]
WebElements = List[WebElement]

WebDriverManagers = Union[
    ChromeDriverManager,
    GeckoDriverManager,
    EdgeChromiumDriverManager
]

class TypeBrowser(TypedDict):
    versions: List[int]
    default: int
    kwargs: Dict[str, Any]



import gc
import logging
import traceback
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Dict
from pathlib import Path

from bs4 import BeautifulSoup
from lxml import etree


from weberist.generic.shortcuts import WebDriverWait
from weberist.generic.shortcuts import expected_conditions as EC
from weberist.generic.utils import extract_base_url
from weberist.generic.constants import ATTR_SELECTOR
from weberist.generic.types import (
    WebDriver,
    WebElement,
    WebElements,
    WebDriverServices,
)
from weberist.utils.helpers import Key
from weberist.utils.javascript import (
    document_query_selector,
    document_query_selector_all,
    is_display,
    DISPATCH_ENTER,
    DISPATCH_ENTER_SELECTOR,
)
from .data import ProfileStorageBackend
from .managers import WebDriverFactory
from .exceptions import (
    EXCEPTIONS,
    WebDriverException,
    NoSuchWindowException,
    NoSuchElementException,
    InvalidSessionIdException,
)


logger = logging.getLogger('base.drivers')
logger.setLevel(logging.DEBUG)


class BaseDriver(WebDriverFactory):
    
    def __new__(cls,
                *args,
                browser: str = 'chrome',
                option_arguments: List[str] = None,
                services_kwargs: dict[str, Any] = None,
                keep_alive: bool = True,
                extensions: List[str | Path] = None,
                capabilities: Dict = None,
                quit_on_failure: bool = False,
                timeout: int = 20,
                profile: str = None,
                localstorage: str = None,
                **kwargs,) -> WebDriver:
        
        kwargs['profile'] = profile
        kwargs['localstorage'] = localstorage
        
        instance: WebDriver = super().__new__(
            cls,
            *args,
            browser=browser,
            option_arguments=option_arguments,
            services_kwargs=services_kwargs,
            keep_alive=keep_alive,
            extensions=extensions,
            capabilities=capabilities,
            **kwargs,
        )
        
        profile = kwargs['profile'] = kwargs.get('profile', None)
        localstorage = kwargs['localstorage'] = kwargs.get('localstorage', None)
        cls.__init__(
            instance,
            quit_on_failure=quit_on_failure,
            timeout=timeout,
            profile=profile,
            localstorage=localstorage,
        )
        return instance


    def __init__(self: WebDriver,
                 quit_on_failure: bool = False,
                 timeout: int = 20,
                 profile: str = None,
                 localstorage: str = None) -> None:
        
        self._quit_on_failure = quit_on_failure
        self.timeout = timeout
        self.target_path = Path('.')
        self.soup = None
        self.dom = None
        
        if profile and localstorage:
            self.target_path = Path(localstorage)
            self.profile_backend = ProfileStorageBackend(self.target_path)

    def __enter__(self):
        self._quit_on_failure = False
        return self

    def __exit__(self, exc_type, exc_value, traceback_):
        if exc_type or exc_value or traceback_:
            logger.error("Exception occurred: %s", exc_value)
        self.quit_driver()
    
    @property
    def quit_on_failure(self,):
        return self._quit_on_failure

    @staticmethod
    def quitonfailure(method: Callable) -> Callable:
        """
        A decorator that ensures the web driver is safely quit to avoid memory
        leakages in case of exceptions. This method is designed to be used as a
        decorator for methods that may raise exceptions during web automation
        tasks.

        Parameters
        ----------
        method : Callable
            The method to be decorated. This method should be a callable that
            takes an instance of the `BaseCrawler` class as its first argument.

        Returns
        -------
        Callable
            The decorated method, which will now include additional error
            handling to quit the
            web driver if an exception is raised.

        Notes
        -----
        This method is a static method and should be used as a decorator.
        """
        # pylint: disable=W0613
        def inner(self, *args, **kwargs) -> Callable:
            try:
                # pylint: disable=not-callable
                return method(self, *args, **kwargs)
            except EXCEPTIONS as err:
                logger.error(err)
                logger.error(traceback.format_exc())
                if self.quit_on_failure:
                    logger.warning("Closing window and quitting driver.")
                    self.quit_driver()
                    logger.warning("Driver quit.")
                raise err

        return inner

    def quit_driver(self):
        try:
            self.quit()
        except EXCEPTIONS as err:
            logger.error("Error while quitting driver: %s", err)
        finally:
            logger.info("Driver quit successfully.")
            self._cleanup()

    def _cleanup(self):
        if hasattr(self, 'service') and self.service:
            try:
                self.service.stop()
            except EXCEPTIONS as err:
                logger.error("Error while stopping service: %s", err)
        gc.collect()

    def is_running(self,) -> bool:
        """
        Checks if the web driver is currently running.

        This method attempts to assert that the web driver's service process
        is still running. It returns True if the process is running, and False
        otherwise.

        Returns
        -------
        bool
            True if the web driver's service process is running, False
            otherwise.
        """
        try:
            return isinstance(self.current_url, str)
        except (WebDriverException, InvalidSessionIdException):
            return False

    @quitonfailure
    def base_url(self,):
        return extract_base_url(self.current_url)

    @quitonfailure
    def tab_handle(self):
        """
        Retrieves the handle of the current window (tab) of the web driver.

        Returns
        -------
        str
            The handle of the current window (tab) of the web driver.
        """
        return self.current_window_handle

    @quitonfailure
    def tabs(self):
        """
        Retrieves a list of handles for all open windows (tabs) of the web
        driver.

        Returns
        -------
        List[str]
            A list of handles for all open windows (tabs) of the web driver.
        """
        return self.window_handles

    @quitonfailure
    def switch_to_tab(self, index: int):
        """
        Switches the web driver's focus to the tab at the specified index.

        Parameters
        ----------
        index : int
            The index of the tab to switch to.

        Raises
        ------
        IndexError
            If the index is out of range of the available tabs.
        """
        tabs = self.tabs()
        if len(tabs) <= index:
            logger.warning(
                "Trying to switch to tab nº %d, but there are only %d tabs",
                index,
                len(tabs),
            )
        else:
            self.switch_to.window(tabs[index])

    def wait(
        self,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException, )
    ):
        """
        Returns a WebDriverWait object that can be used to wait for a condition
        to be met.

        Parameters
        ----------
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebDriverWait
            A WebDriverWait object configured with the specified parameters.
        """
        if timeout is None:
            timeout = self.timeout
        return WebDriverWait(
            self, timeout, poll_frequency, ignored_exceptions
        )

    @quitonfailure
    def close_tab(self):
        """
        Attempts to close the current tab of the web driver. If the current
        tab cannot be closed, it switches to the first tab and closes it.

        Raises
        ------
        NoSuchWindowException
            If the current tab cannot be closed.
        """
        try:
            self.close()
        except NoSuchWindowException:
            tab_index = 0
            current_tab = self.tab_handle()
            for index, tab in enumerate(self.tabs()):
                if tab == current_tab:
                    tab_index = index
            logger.warning(
                "Tab %d no longer exists. Switching to tab 0", tab_index
            )
            self.switch_to_tab(0)

    def close_driver(self, quit_driver=True):
        """
        Attempts to close the current window of the web driver. If the current
        window cannot be closed, it switches to the first tab and closes it.
        If `quit_driver` is True, it also quits the web driver.

        Parameters
        ----------
        quit_driver : bool, optional
            Whether to quit the web driver after closing the current window.
            Default is True.
        """
        try:
            self.close()
        except NoSuchWindowException:
            self.switch_to_tab(0)
            self.close()
        except WebDriverException:
            pass
        if quit_driver:
            self.quit()
    
    @quitonfailure
    def goto(self, url: str) -> None:
        """
        Navigates the web driver to the specified URL.

        Parameters
        ----------
        url : str
            The URL to navigate to.
        """
        self.get(url)

    @quitonfailure
    def select(
        self,
        value: str,
        by: str = "id",
        expected_condition=EC.presence_of_element_located,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ) -> WebElement:
        """
        Finds an element on the web page using the specified locator strategy
        and waits until the element is present.

        Parameters
        ----------
        value : str
            The value to search for, such as the ID, name, or XPath of the
            element.
        by : str, default "id"
            The locator strategy to use, such as "id", "name", "xpath", etc.
        expected_condition : Callable, default EC.presence_of_element_located
            The expected condition to wait for, such as the presence of the
            element.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElement
            The first matching element found.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        if timeout is None:
            timeout = self.timeout
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        result = wait.until(expected_condition((ATTR_SELECTOR[by], value)))
        return result

    @quitonfailure
    def select_elements(
        self,
        value: str,
        by: str = "id",
        expected_condition=EC.presence_of_all_elements_located,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ) -> WebElements:
        """
        Finds all elements on the web page that match the specified locator
        strategy and waits until at least one element is present.


        Parameters
        ----------
        value : str
            The value to search for, such as the ID, name, or XPath of the
            elements.
        by : str, default "id"
            The locator strategy to use, such as "id", "name", "xpath", etc.
        expected_condition : Callable, default
        EC.presence_of_all_elements_located
            The expected condition to wait for, such as the presence of the
            elements.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElements
            A list of all matching elements found.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        return self.select(
            value,
            by,
            expected_condition,
            timeout,
            poll_frequency,
            ignored_exceptions
        )

    @quitonfailure
    def xpath(
        self,
        value: str,
        expected_condition=EC.presence_of_all_elements_located,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ) -> WebElements:
        """
        Finds all elements on the web page that match the specified XPath and
        waits until at least one element is present.


        Parameters
        ----------
        value : str
            The XPath expression to search for.
        expected_condition : Callable, default EC.presence_of_all_elements_located  # noqa E501
            The expected condition to wait for, such as the presence of the
            elements.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElements
            A list of all matching elements found.
        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        return self.select_elements(
            value,
            "xpath",
            expected_condition,
            timeout,
            poll_frequency,
            ignored_exceptions
        )

    @quitonfailure
    def find_text(self,
                    value: str,
                    tag: str = None,
                    expected_condition=EC.presence_of_all_elements_located,
                    timeout=None,
                    poll_frequency=0.5,
                    ignored_exceptions=(NoSuchElementException,)):
        if tag is None:
            tag = "*"
        xpath = f"//{tag}[text()='{value}']"
        result = self.xpath(
            value=xpath,
            expected_condition=expected_condition,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions,
        )
        if result and len(result) > 0:
            return result
        return None

    @quitonfailure
    def find_contains_text(self,
                           value: str,
                           tag: str = None,
                           expected_condition=EC.presence_of_all_elements_located,
                           timeout=None,
                           poll_frequency=0.5,
                           ignored_exceptions=(NoSuchElementException,)):
        if tag is None:
            tag = "*"
        xpath = f"//{tag}[contains(text(), '{value}')]"
        result = self.xpath(
            value=xpath,
            expected_condition=expected_condition,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions,
        )
        if result and len(result) > 0:
            return result
        return None
    
    @quitonfailure
    def send_to(
        self,
        element: WebElement,
        key: str | Key,
        expected_condition=EC.element_to_be_clickable,
        enter=False,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Sends a key to the specified web element and optionally presses the
        Enter key.


        Parameters
        ----------
        element : WebElement
            The web element to send the key to.
        key : str | Key
            The key or text to send to the element.
        expected_condition : Callable, default EC.element_to_be_clickable
            The expected condition to wait for before sending the key.
        enter : bool, default False
            Whether to press the Enter key after sending the key.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElement
            The element to which the key was sent.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        wait.until(expected_condition(element)).send_keys(key)
        if enter:
            element.send_keys(Key.enter)

    @quitonfailure
    def send(
        self,
        value: str,
        key: str | Key,
        by="id",
        expected_condition_element=EC.presence_of_element_located,
        expected_condition_send=EC.element_to_be_clickable,
        enter=False,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        element = self.select(
            value=value,
            by=by,
            expected_condition=expected_condition_element,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions
        )
        self.send_to(
            element,
            key,
            expected_condition=expected_condition_send,
            enter=enter,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions
        )

    @quitonfailure
    def child(
        self,
        element: WebElement,
        value: str,
        by="id",
        expected_condition=EC.visibility_of,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Finds a child element of the specified web element using the specified
        locator strategy and waits until the element is visible.

        Parameters
        ----------
        element : WebElement
            The parent web element to find the child element within.
        value : str
            The value to search for, such as the ID, name, or XPath of the
            child element.
        by : str, default "id"
            The locator strategy to use, such as "id", "name", "xpath", etc.
        expected_condition : Callable, default EC.visibility_of
            The expected condition to wait for, such as the visibility of the
            child element.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElement
            The first matching child element found.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        wait.until(expected_condition(element))
        descendant = wait.until(
            lambda elem: element.find_element(ATTR_SELECTOR[by], value)
        )
        return descendant

    @quitonfailure
    def child_by_class_name(
        self,
        element: WebElement,
        value: str,
        expected_condition=EC.visibility_of,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Finds a child element of the specified web element by class name and
        waits until the element is visible.

        Parameters
        ----------
        element : WebElement
            The parent web element to find the child element within.
        value : str
            The class name to search for.
        expected_condition : Callable, default EC.visibility_of
            The expected condition to wait for, such as the visibility of the
            child element.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElement
            The first matching child element found by class name.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        wait.until(expected_condition(element))
        descendant = wait.until(
            lambda elem: element.find_element(
                ATTR_SELECTOR['class name'], value)
        )
        return descendant

    @quitonfailure
    def children(
        self,
        element: WebElement,
        value: str,
        by="id",
        expected_condition=EC.visibility_of,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Finds all child elements of the specified web element using the
        specified locator strategy and waits until at least one element is
        visible.

        Parameters
        ----------
        element : WebElement
            The parent web element to find the child elements within.
        value : str
            The value to search for, such as the ID, name, or XPath of the
            child elements.
        by : str, default "id"
            The locator strategy to use, such as "id", "name", "xpath", etc.
        expected_condition : Callable, default EC.visibility_of
            The expected condition to wait for, such as the visibility of the
            child elements.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElements
            A list of all matching child elements found.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        wait.until(expected_condition(element))
        offspring = wait.until(
            lambda elem: element.find_elements(ATTR_SELECTOR[by], value)
        )
        return offspring

    @quitonfailure
    def children_by_class_name(
        self,
        element: WebElement,
        value: str,
        expected_condition=EC.visibility_of,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Finds all child elements of the specified web element by class name and
        waits until at least one element is visible.

        Parameters
        ----------
        element : WebElement
            The parent web element to find the child elements within.
        value : str
            The class name to search for.
        expected_condition : Callable, default EC.visibility_of
            The expected condition to wait for, such as the visibility of the
            child elements.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Returns
        -------
        WebElements
            A list of all matching child elements found by class name.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        wait.until(expected_condition(element))
        offspring = wait.until(
            lambda elem: element.find_elements(
                ATTR_SELECTOR['class name'], value)
        )
        return offspring

    @quitonfailure
    def click_element(
        self,
        element: WebElement,
        expected_condition=EC.element_to_be_clickable,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Clicks on the specified web element after waiting for it to be
        clickable.

        Parameters
        ----------
        element : WebElement
            The web element to click on.
        expected_condition : Callable, default EC.element_to_be_clickable
            The expected condition to wait for before clicking the element.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        wait.until(expected_condition(element)).click()

    @quitonfailure
    def click(
        self,
        value: str,
        by="id",
        expected_condition_element=EC.presence_of_element_located,
        expected_condition_click=EC.element_to_be_clickable,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Finds an element on the web page using the specified locator strategy
        and waits until the element is present, then clicks on it.

        Parameters
        ----------
        value : str
            The value to search for, such as the ID, name, or XPath of the
            element.
        by : str, default "id"
            The locator strategy to use, such as "id", "name", "xpath", etc.
        expected_condition_element : Callable, default EC.presence_of_element_located  #noqa E501
            The expected condition to wait for, such as the presence of the
            element.
        expected_condition_click : Callable, default EC.element_to_be_clickable
            The expected condition to wait for before clicking the element.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        element = self.select(
            value=value,
            by=by,
            expected_condition=expected_condition_element,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions
        )
        self.click_element(
            element,
            expected_condition=expected_condition_click,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions
        )

    @quitonfailure
    def arrow_down_element(
        self,
        element: WebElement,
        n_times: int = 1,
        enter=False,
        expected_condition=EC.element_to_be_clickable,
        timeout=None,
        poll_frequency=0.5,
        ignored_exceptions=(NoSuchElementException,)
    ):
        """
        Sends the down arrow key to the specified web element the specified
        number of times and optionally presses the Enter key.

        Parameters
        ----------
        element : WebElement
            The web element to send the down arrow key to.
        n_times : int, default 1
            The number of times to send the down arrow key.
        enter : bool, default False
            Whether to press the Enter key after sending the down arrow key.
        expected_condition : Callable, default EC.element_to_be_clickable
            The expected condition to wait for before sending the keys.
        timeout : int, optional
            The maximum time to wait for the condition to be met. If not
            specified, the default timeout is used.
        poll_frequency : float, default 0.5
            The frequency at which the condition is checked, in seconds.
        ignored_exceptions : tuple, default (NoSuchElementException,)
            Exceptions to ignore while waiting for the condition.

        Raises
        ------
        TimeoutException
            If the condition is not met within the specified timeout.
        """
        wait = self.wait(timeout, poll_frequency, ignored_exceptions)
        for _ in range(n_times):
            wait.until(expected_condition(element)).send_keys(Key.down)
        if enter:
            wait.until(expected_condition(element)).send_keys(Key.enter)

    @quitonfailure
    def soup_of(
        self,
        element: WebElement,
        parser="html.parser",
        features="lxml",
        outer=True,
        **kwargs
    ):
        """
        Parses the inner or outer HTML of the specified web element using
        BeautifulSoup.

        Parameters
        ----------
        element : WebElement
            The web element to parse.
        parser : str, default "html.parser"
            The parser to use for parsing the HTML.
        features : str, default "lxml"
            The features to use for parsing the HTML.
        outer : bool, default True
            Whether to parse the outer HTML of the element. If False, the inner
            HTML is parsed.
        **kwargs : dict
            Additional keyword arguments to pass to the BeautifulSoup
            constructor.

        Returns
        -------
        BeautifulSoup
            A BeautifulSoup object representing the parsed HTML of the element.
        """
        type_attribute = "innerHTML"
        if outer:
            type_attribute = "outerHTML"

        return BeautifulSoup(
            element.get_attribute(type_attribute),
            parser=parser,
            features=features,
            **kwargs
        )

    @quitonfailure
    def run(self, script, *args):
        """
        Executes the specified JavaScript script on the web page.

        Parameters
        ----------
        script : str
            The JavaScript script to execute.
        *args : tuple
            Arguments to pass to the JavaScript script.

        Returns
        -------
        Any
            The result of the script execution.
        """
        return self.execute_script(script, *args)

    @quitonfailure
    def query_selector(self, selector: str):
        """
        Executes a JavaScript query selector on the web page and returns the
        first matching element.

        Parameters
        ----------
        selector : str
            The CSS selector to use for querying the web page.

        Returns
        -------
        WebElement
            The first matching element found by the query selector.
        """
        script = document_query_selector(selector)
        return self.run(script)

    @quitonfailure
    def query_selector_all(self, selector: str):
        """
        Executes a JavaScript query selector on the web page and returns all
        matching elements.

        Parameters
        ----------
        selector : str
            The CSS selector to use for querying the web page.

        Returns
        -------
        WebElements
            A list of all matching elements found by the query selector.
        """
        script = document_query_selector_all(selector)
        return self.run(script)

    @quitonfailure
    def dispatch_enter(self, element: WebElement):
        """
        Dispatches an 'Enter' key event to the specified web element.

        Parameters
        ----------
        element : WebElement
            The web element to dispatch the 'Enter' key event to.

        Returns
        -------
        Any
            The result of the dispatched event.
        """
        return self.run(DISPATCH_ENTER, element)

    @quitonfailure
    def dispatch_enter_selector(self, selector: str):
        """
        Dispatches an 'Enter' key event to the first element matching the
        specified CSS selector.

        Parameters
        ----------
        selector : str
            The CSS selector to use for querying the web page.

        Returns
        -------
        Any
            The result of the dispatched event.
        """
        return self.run(DISPATCH_ENTER_SELECTOR.format(selector))

    @quitonfailure
    def make_soup(self, parser="html.parser", **kwargs):
        """
        Parses the current page source using BeautifulSoup with the specified
        parser.

        Parameters
        ----------
        parser : str, default "html.parser"
            The parser to use for parsing the HTML.
        **kwargs : dict
            Additional keyword arguments to pass to the BeautifulSoup
            constructor.

        Returns
        -------
        BeautifulSoup
            A BeautifulSoup object representing the parsed HTML of the current
            page.
        """
        return BeautifulSoup(self.page_source(), parser=parser, **kwargs)

    @quitonfailure
    def make_dom(self, soup_parser="html.parser", **kwargs):
        """
        Parses the current page source using lxml with the specified parser and
        returns an ElementTree object.

        Parameters
        ----------
        soup_parser : str, default "html.parser"
            The parser to use for parsing the HTML.
        **kwargs : dict
            Additional keyword arguments to pass to the lxml HTML constructor.

        Returns
        -------
        etree.ElementTree
            An ElementTree object representing the parsed HTML of the current
            page.
        """
        self.soup = self.make_soup(parser=soup_parser)
        self.dom = etree.HTML(str(self.soup), **kwargs)
        return self.dom

    def is_display(self, element: WebElement, value: str):
        """
        Checks if the specified web element is displayed with the given display
        value.

        Parameters
        ----------
        element : WebElement
            The web element to check the display property of.
        value : str
            The display value to check against, such as "block", "none", etc.

        Returns
        -------
        bool
            True if the element is displayed with the given value, False
            otherwise.
        """
        script = is_display(value)
        return self.run(script, element)

    @quitonfailure
    def switch_to_frame(self, value: str, by='id'):
        """
        Switches the web driver's focus to the frame specified by the given
        value and locator strategy.

        Parameters
        ----------
        value : str
            The value to search for, such as the ID, name, or XPath of the frame.  # noqa E501
        by : str, default 'id'
            The locator strategy to use, such as "id", "name", "xpath", etc.

        Raises
        ------
        TimeoutException
            If the frame is not found within the specified timeout.
        """
        self.wait().until(
            EC.frame_to_be_available_and_switch_to_it(
                (by, value)
            )
        )



"""
    shelob.core.adapters



    This module should provide adapter functions for Crawlers objects.
    # TODO:
        - add more javascript adapters;
        - complete documentation;
"""

import logging

from weberist.generic.types import WebElement
from weberist.generic.constants import DISPLAY_VALUES


logger = logging.getLogger('standard')


DISPATCH_ENTER = """var ke = new KeyboardEvent('keydown', {
    bubbles: true, cancelable: true, keyCode: 13
});
arguments[0].dispatchEvent(ke);
"""
DISPATCH_ENTER_SELECTOR = (
    "var ke = new KeyboardEvent('keydown', {{"
    "   bubbles: true, cancelable: true, keyCode: 13"
    "}});"
    "{}.dispatchEvent(ke);"
)




def selector(value, by) -> dict[str, str]:
    return {
        "value": value,
        "by": by
    }

def document_query_selector(selector_str: str) -> str:
    """String to parse javascript code."""
    return f"return document.querySelector('{selector_str}');"


def document_query_selector_all(selector_str: str) -> str:
    """String to parse javascript code."""
    return f"return document.querySelectorAll('{selector_str}');"


def document_query_selector_click(selector_str: str) -> str:
    """String to parse javascript code."""
    element = document_query_selector(selector_str=selector_str)
    return (
        f"const element = {element};"
        "element.click();"
        "return element;"
    )

def is_display(
        value: str, element: WebElement = None
) -> str | tuple[str, WebElement]:
    if value not in DISPLAY_VALUES:
        raise ValueError(
            f'Invalid display value "{value}". Valid are {DISPLAY_VALUES}.'
        )
    script = f"return arguments[0].style.display == '{value}';"
    if WebElement is None:
        return script
    return script, element



from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By

# lower casing
class Key(Keys):  # pylint: disable=too-few-public-methods
    """ "keys to use in send functions"""

    enter = Keys.RETURN
    esc = Keys.ESCAPE
    delete = Keys.DELETE
    down = Keys.ARROW_DOWN
    up = Keys.ARROW_UP
    tab = Keys.TAB
    backspace = Keys.BACK_SPACE
    shift = Keys.SHIFT
    left_shift = Keys.LEFT_SHIFT
    control = Keys.CONTROL
    left_control = Keys.LEFT_CONTROL
    alt = Keys.ALT
    left_alt = Keys.LEFT_ALT
    pause = Keys.PAUSE
    space = Keys.SPACE
    page_up = Keys.PAGE_UP
    page_down = Keys.PAGE_DOWN
    end = Keys.END
    home = Keys.HOME
    left = Keys.LEFT
    right = Keys.RIGHT
    insert = Keys.INSERT
    semicolon = Keys.SEMICOLON
    equals = Keys.EQUALS
    numpad0 = Keys.NUMPAD0
    numpad1 = Keys.NUMPAD1
    numpad2 = Keys.NUMPAD2
    numpad3 = Keys.NUMPAD3
    numpad4 = Keys.NUMPAD4
    numpad5 = Keys.NUMPAD5
    numpad6 = Keys.NUMPAD6
    numpad7 = Keys.NUMPAD7
    numpad8 = Keys.NUMPAD8
    numpad9 = Keys.NUMPAD9
    multiply = Keys.MULTIPLY
    add = Keys.ADD
    separator = Keys.SEPARATOR
    subtract = Keys.SUBTRACT
    decimal = Keys.DECIMAL
    divide = Keys.DIVIDE
    f1 = Keys.F1
    f2 = Keys.F2
    f3 = Keys.F3
    f4 = Keys.F4
    f5 = Keys.F5
    f6 = Keys.F6
    f7 = Keys.F7
    f8 = Keys.F8
    f9 = Keys.F9
    f10 = Keys.F10
    f11 = Keys.F11
    f12 = Keys.F12
    meta = Keys.META
    command = Keys.COMMAND
    zenkaku_hankaku = Keys.ZENKAKU_HANKAKU


class SelectorType(By):
    """Set of supported locator strategies."""

    id = By.ID
    xpath = By.XPATH
    link_text = By.LINK_TEXT
    partial_link_text = By.PARTIAL_LINK_TEXT
    name = By.NAME
    tag_name = By.TAG_NAME
    class_name = By.CLASS_NAME
    css_selector = By.CSS_SELECTOR



"""
This module provides a centralized management system for different web drivers,
including Firefox, Chrome, Safari, and Edge. It allows for easy access to the
web drivers, their options, services, and managers, facilitating the creation
and configuration of web driver instances for automated testing or web scraping
tasks.

The module includes a class `WebDrivers` that encapsulates the web drivers and
their associated components. It provides properties to access each web driver,
its options, services, and managers, as well as a method to retrieve a tuple
containing the web driver, its options, services, and managers for a specified
browser.

This module is part of the weberist project, a web automation and scraping
framework.
"""
import gc
import time
import socket
import logging
from re import match
from typing import Any, List, Dict, Tuple
from pathlib import Path

from selenium.webdriver.common.options import BaseOptions
from selenium.webdriver.remote.file_detector import FileDetector
from selenium.webdriver.remote.remote_connection import RemoteConnection
from selenium_stealth import (
    with_utils,
    chrome_app,
    chrome_runtime,
    iframe_content_window,
    media_codecs,
    navigator_languages,
    navigator_permissions,
    navigator_plugins,
    navigator_vendor,
    navigator_webdriver,
    user_agent_override,
    webgl_vendor_override,
    window_outerdimensions,
    hairline_fix,
)
from selenium_stealth.wrapper import evaluateOnNewDocument

from weberist.generic.shortcuts import (
    Firefox,
    Chrome,
    Safari,
    Edge,
    FirefoxOptions,
    ChromeOptions,
    SafariOptions,
    EdgeOptions,
    FirefoxService,
    ChromeService,
    SafariService,
    EdgeService,
    ChromeDriverManager,
    GeckoDriverManager,
    EdgeChromiumDriverManager,
    SeleniumWebDriver,
)

from weberist.generic.types import (
    WebDriver,
    WebDriverOptions,
    WebDriverServices,
    WebDriverManagers
)
from weberist.generic.constants import (
    DEFAULT_ARGUMENTS,
    SELENOID_CAPABILITIES,
    SUPPORTED_BROWSERS,
)

from .data import UserAgent, WindowSize
from .config import DEFAULT_PROFILE, ROOT_DIR, LOCALSTORAGE
from .stealth.tools import remove_cdc

logger = logging.getLogger('standard')

def free_port() -> int:
    """Get free port."""
    sock = socket.socket()
    sock.bind(('localhost', 0))
    port = sock.getsockname()[1]
    sock.close()
    del sock
    gc.collect()
    return port

def add_option(option: WebDriverOptions, arguments, browser: str = 'chrome'):

    proceed_argument = True
    proceed_experimental = True
    for argument in arguments:
        if isinstance(argument, str) and proceed_argument:
            try:
                getattr(option, 'add_argument')(argument)
            except AttributeError:
                logger.warning("'%s' doesn't support adding option", browser)
                proceed_argument = False
                continue
        if isinstance(argument, dict):
            if 'chrome' in browser and proceed_experimental:
                for name, value in argument.items():
                    try:
                        getattr(option, 'add_experimental_option')(name, value)
                    except AttributeError:
                        logger.warning(
                            "'%s' doesn't support adding experimental option",
                            browser
                        )
                        proceed_argument = False
                        break
        if not (proceed_argument or proceed_experimental):
            return None
    return option


class WebDrivers:
    """
    A class that encapsulates the management of web drivers for different
    browsers, including Firefox, Chrome, Safari, and Edge. It provides
    properties to access each web driver, its options, services, and managers.
    This class is designed to simplify the process of configuring and
    retrieving web driver instances for automated testing or web scraping
    tasks.

    Attributes
    ----------
    firefox : Firefox
        An instance of the Firefox web driver.
    chrome : Chrome
        An instance of the Chrome web driver.
    safari : Safari
        An instance of the Safari web driver.
    edge : Edge
        An instance of the Edge web driver.
    firefox_options : FirefoxOptions
        Options for the Firefox web driver.
    chrome_options : ChromeOptions
        Options for the Chrome web driver.
    safari_options : SafariOptions
        Options for the Safari web driver.
    edge_options : EdgeOptions
        Options for the Edge web driver.
    firefox_service : FirefoxService
        Service for the Firefox web driver.
    chrome_service : ChromeService
        Service for the Chrome web driver.
    safari_service : SafariService
        Service for the Safari web driver.
    edge_service : EdgeService
        Service for the Edge web driver.
    firefox_manager : GeckoDriverManager
        Manager for the Firefox web driver.
    chrome_manager : ChromeDriverManager
        Manager for the Chrome web driver.
    safari_manager : None
        Manager for the Safari web driver.
    edge_manager : EdgeChromiumDriverManager
        Manager for the Edge web driver.
    supported : tuple[str]
        A tuple of supported browser names.
    """
    __drivers = {
        "firefox": {
            "driver": Firefox,
            "options": FirefoxOptions,
            "service": FirefoxService,
            "manager": GeckoDriverManager
        },
        "chrome": {
            "driver": Chrome,
            "options": ChromeOptions,
            "service": ChromeService,
            "manager": ChromeDriverManager
        },
        "chrome_remote": {
            "driver": SeleniumWebDriver,
            "options": ChromeOptions,
            "service": None,
            "manager": None
        },
        "safari": {
            "driver": Safari,
            "options": SafariOptions,
            "service": SafariService,
            "manager": None
        },
        "edge": {
            "driver": Edge,
            "options": EdgeOptions,
            "service": EdgeService,
            "manager": EdgeChromiumDriverManager
        }
    }

    supported: Tuple[str] = tuple(__drivers.keys())

    def get(self,
            browser: str,
            option_arguments: List[str] = None,
            extensions: str | List[str] = None,
            capabilities: Dict = None,
            services_kwargs: Dict = None,
            **kwargs) -> tuple[
        WebDriver, WebDriverOptions, WebDriverServices, WebDriverManagers
    ]:
        """
        Retrieves a tuple containing the web driver, its options, services,
        and managers for the specified browser.

        This method dynamically accesses the appropriate attributes based on
        the browser name provided. It raises an AttributeError if the specified
        browser is not supported.

        Parameters
        ----------
        browser : str
            The name of the browser for which to retrieve the web driver, its
            options, services, and managers. Supported browsers include
            "firefox", "chrome", "safari" and "edge".

        Returns
        -------
        tuple[WebDriver, WebDriverOptions, WebDriverServices, WebDriverManagers]  # noqa E501
            A tuple containing the web driver, its options, services, and
            managers for the specified browser.

        Raises
        ------
        AttributeError
            If the specified browser is not supported.
        """
        if browser not in self.supported:
            raise AttributeError(
                f'WebDrivers does not support driver for {browser}'
            )

        driver = self.__drivers[browser]['driver']
        option, service = self._configure(
            browser,
            option_arguments,
            extensions,
            capabilities,
            services_kwargs,
            **kwargs
        )
        if 'remote' in browser:
            browser = browser.split('_')[0]
        return driver, option, service

    def _configure_chrome(self,
                          options: WebDriverOptions,
                          manager: WebDriverManagers,
                          service_class: type[WebDriverServices],
                          option_arguments: List[str],
                          extensions: str | List[str] = None,
                          capabilities: Dict = None,
                          service_kwargs: Dict = None,
                          host: str = None,
                          port: int = None,
                          lang: str = 'en-US',
                          remote: bool = False):

        if extensions:
            if all(isinstance(item, Path) for item in extensions):
                argument = '--load-extension=' + ','.join(
                    [str(path) for path in extensions]
                )
                options.add_argument(argument)
            else:
                for argument in extensions:
                    options.add_extension(argument)

        user_agent_string = None
        windows_size_ = None
        profile_name = None

        for argument in option_arguments:

            if user_agent_string is None and 'user-agent' in argument:
                user_agent_string = argument.split("=")[-1]
                continue
            if windows_size_ is None and 'windows-size' in argument:
                windows_size_ = argument.split("=")[-1]
                continue
            if profile_name is None and 'profile-directory' in argument:
                profile_name = argument.split("=")[-1]

        user_agent = UserAgent()
        windows_size = WindowSize()

        if profile_name is not None:
            user_agent_string = user_agent.get_hashed(profile_name)
            windows_size_ = windows_size.get_hashed(profile_name)
        if user_agent_string is None:
            user_agent_string = user_agent.get_random()
        if windows_size_ is None:
            windows_size_ = windows_size.get_random()
        windows_size_string = windows_size_
        if not isinstance(windows_size_, str):
            windows_size_string = windows_size.to_string(windows_size_)
        option_arguments.extend(
            [
                f"--user-agent={user_agent_string}",
                f"--window-size={windows_size_string}"
            ]
        )

        host = host or "127.0.0.1"
        port = port or free_port()
        option_arguments.append(f"--remote-debugging-host={host}")
        option_arguments.append(f"--remote-debugging-port={port}")
        option_arguments.append(f'--lang={lang}')

        service = None
        if capabilities is None:
            capabilities = {}
        if remote:
            capabilities.update(SELENOID_CAPABILITIES)
            for option in options.arguments:
                if '--user-data-dir' in option:
                    capabilities['selenoid:options']['env'] = (
                        [f'BROWSER_PROFILE_DIR={option.split("=")[-1]}']
                    )
                    if profile_name is None:
                        profile_name = DEFAULT_PROFILE
                    break
        else:
            executable_path = None
            if hasattr(manager, 'install'):
                executable_path = manager().install()
                if service_kwargs:
                    service = service_class(executable_path, **service_kwargs)
                else:
                    service = service_class(executable_path)
                remove_cdc(service.path)

        for name, value in capabilities.items():
            options.set_capability(name, value)

        options = add_option(options, option_arguments, 'chrome')

        return options, service

    def _configure(self,
                   browser: str,
                   option_arguments: List[str],
                   extensions: str | List[str] = None,
                   capabilities: Dict = None,
                   service_kwargs: Dict = None,
                   **kwargs):

        browser_name = browser
        remote = False
        if 'remote' in browser:
            remote = True
            browser_name = browser.split('_')[0]

        options_class = self.__drivers[browser_name]['options']
        service_class = self.__drivers[browser_name]['service']
        manager = self.__drivers[browser_name]['manager']

        options: WebDriverOptions = options_class()
        option_arguments = option_arguments or []

        if browser_name == 'chrome':
            return self._configure_chrome(
                options,
                manager,
                service_class=service_class,
                option_arguments=option_arguments,
                extensions=extensions,
                capabilities=capabilities,
                service_kwargs=service_kwargs,
                remote=remote,
                **kwargs,
            )

        if extensions:
            logger.warning("Extensions only implemented for chrome.")
        if capabilities:
            logger.warning("Capabilities only implemented for chrome.")

        options = add_option(options, option_arguments, browser)
        options = options or options_class()

        service = None
        executable_path = None
        if hasattr(manager, 'install'):
            executable_path = manager().install()
            if service_kwargs:
                service = service_class(executable_path, **service_kwargs)
            else:
                service = service_class(executable_path)

        return options, service


class WebDriverFactory(SeleniumWebDriver):
    
    service: WebDriverServices = None
    
    @classmethod
    def _set_up(cls,
                browser: str,
                arguments: List[str | Dict],
                **kwargs):

        cls_properties = {
            name: getattr(cls, name)
            for name in dir(cls) if not match("__.*__", name)
        }
        if browser.split("_")[0] in DEFAULT_ARGUMENTS:
            arguments = arguments or []
            arguments.extend(list(DEFAULT_ARGUMENTS[browser.split("_")[0]]))

        if 'remote' in browser and 'command_executor' not in kwargs:
            kwargs['command_executor'] = "http://0.0.0.0:4444/wd/hub"

        if 'chrome' in browser:
            experimental_options = kwargs.pop("experimental_options", {})
            if 'profile' in kwargs and kwargs['profile']:
                arguments.append(
                    f"--profile-directory={kwargs['profile']}"
                )
                kwargs.pop('profile')
            if 'localstorage' in kwargs and kwargs['localstorage']:
                arguments.append(
                    f"--user-data-dir={kwargs['localstorage']}"
                )
                kwargs.pop('localstorage')
            experimental_options.update(
                {
                    "excludeSwitches": ["enable-automation"],
                    "useAutomationExtension": False
                }
            )
            arguments.append(experimental_options)

        kwargs.pop('quit_on_failure', None)
        kwargs.pop('timeout', None)

        return browser, cls_properties, arguments, kwargs

    def __new__(cls,
                *args,
                browser: str = 'chrome',
                option_arguments: List[str] = None,
                services_kwargs: dict[str, Any] = None,
                keep_alive: bool = True,
                extensions: List[str | Path] = None,
                capabilities: Dict = None,
                stealth: bool = True,
                **kwargs,) -> WebDriver:

        capabilities = kwargs.get('capabilities', None)
        browser, cls_properties, option_arguments, kwargs = cls._set_up(
            browser, option_arguments, **kwargs
        )

        host = kwargs.pop("host", None)
        port = kwargs.pop("port", None)
        lang = kwargs.pop("lang", 'en')
        # selenium-stealth arguments:
        languages = kwargs.pop("languages", ["en-US", "en"])
        vendor = kwargs.pop("vendor", "Google Inc.")
        webgl_vendor = kwargs.pop("webgl_vendor", "Intel Inc.")
        renderer = kwargs.pop("renderer", "Intel Iris OpenGL Engine")
        run_on_insecure_origins = kwargs.pop("run_on_insecure_origins", False)

        driver, options, service = WebDrivers().get(
            browser,
            option_arguments,
            extensions,
            capabilities,
            services_kwargs,
            host=host,
            port=port,
            lang=lang,
        )

        if service is not None:
            kwargs['service'] = service

        kwargs.pop('profile', None)
        kwargs.pop('localstorage', None)

        instance: WebDriver = type(cls.__name__, (driver, ), cls_properties)(
            *args,
            options=options,
            keep_alive=keep_alive,
            **kwargs
        )

        if stealth:
            if 'chrome' not in browser:
                logger.warning('Stealthiness only supported in chrome')
                return instance
            if lang not in languages:
                languages.append(lang)
            ua_languages = ','.join(languages)

            # Default selenium_stealth functions
            with_utils(instance, **kwargs)
            chrome_app(instance, **kwargs)
            chrome_runtime(instance, run_on_insecure_origins, **kwargs)
            iframe_content_window(instance, **kwargs)
            media_codecs(instance, **kwargs)
            navigator_languages(instance, languages, **kwargs)
            navigator_permissions(instance, **kwargs)
            navigator_plugins(instance, **kwargs)
            navigator_vendor(instance, vendor, **kwargs)
            navigator_webdriver(instance, **kwargs)
            user_agent_override(instance, ua_languages=ua_languages, **kwargs)
            webgl_vendor_override(instance, webgl_vendor, renderer, **kwargs)
            window_outerdimensions(instance, **kwargs)
            hairline_fix(instance, **kwargs)

            # Hide selenium fingerprints
            selenium_fingerprint = Path(
                ROOT_DIR / 'base/stealth/js/selenium.fingerprint.js'
            ).read_text(encoding='utf-8')
            evaluateOnNewDocument(instance, selenium_fingerprint)

            error_stack_override = Path(
                ROOT_DIR / 'base/stealth/js/error.stack.override.js'
            ).read_text(encoding='utf-8')
            evaluateOnNewDocument(instance, error_stack_override)
            webgl_override = Path(
                ROOT_DIR / 'base/stealth/js/webgl.worker.override.js'
            ).read_text(encoding='utf-8')
            evaluateOnNewDocument(
                instance, webgl_override
            )
            #NOTE: add time.sleep
            # time.sleep(0.5)

        return instance



import sys

from weberist.utils.helpers import SelectorType
from weberist.base.config import CHROME_VERSIONS, FIREFOX_VERSIONS

OPERATING_SYSTEM = sys.platform

DISPLAY_VALUES = (
    'none',
    'inline',
    'block',
    'inline-block'
)

# used in arguments of WeElement selectors
ATTR_SELECTOR = {
    "id": SelectorType.ID,
    "name": SelectorType.NAME,
    "xpath": SelectorType.XPATH,
    "tag name": SelectorType.TAG_NAME,
    "link text": SelectorType.LINK_TEXT,
    "class name": SelectorType.CLASS_NAME,
    "css selector": SelectorType.CSS_SELECTOR,
    "partial link text": SelectorType.PARTIAL_LINK_TEXT,
}

SUPPORTED_BROWSERS: tuple[str] = (
    "chrome",
    "firefox",
    "safari",
    "edge"
)

BROWSERS_VERSIONS = {
    SUPPORTED_BROWSERS[0]: CHROME_VERSIONS,
    SUPPORTED_BROWSERS[1]: FIREFOX_VERSIONS,
}

SELENOID_CAPABILITIES = {
    "browserName": "chrome",
    "browserVersion": f"chrome_{CHROME_VERSIONS[-1]}.0",
    "selenoid:options": {
        "enableVideo": False,
        "enableVNC": True,
    }
}

DEFAULT_ARGUMENTS = {
    SUPPORTED_BROWSERS[0]: (
        "--start-maximized",
        "--no-first-run",
        "--disable-site-isolation-trials",
        "--disable-blink-features=AutomationControlled",
        "--disable-backgrounding-occluded-windows",
        "--disable-hang-monitor",
        "--metrics-recording-only",
        "--disable-sync",
        "--disable-background-timer-throttling",
        "--disable-prompt-on-repost",
        "--disable-background-networking",
        "--disable-infobars",
        "--remote-allow-origins=*",
        "--homepage=about:blank",
        "--no-service-autorun",
        "--disable-ipc-flooding-protection",
        "--disable-session-crashed-bubble",
        "--force-fieldtrials=*BackgroundTracing/default/",
        "--disable-breakpad",
        "--password-store=basic",
        "--disable-features=IsolateOrigins,site-per-process",
        "--disable-features=PrivacySandboxSettings4",
        "--disable-client-side-phishing-detection",
        "--use-mock-keychain",
        "--no-pings",
        "--no-sandbox",
        "--disable-gpu",
        "--disable-popup-blocking",
        "--disable-renderer-backgrounding",
        "--disable-component-update",
        "--disable-dev-shm-usage",
        "--disable-default-apps",
        "--disable-domain-reliability",
        "--no-default-browser-check",
    ),
}

LANGUAGES = {
    'en': ["en-US", "en"],
    'pt': ["pt-BR", "pt"],
}



import logging
from typing import List, Any, Dict
from pathlib import Path

from weberist.base.drivers import BaseDriver
from weberist.base.config import LOCALSTORAGE
from weberist.generic.types import WebDriver

from weberist.base.data import ProfileStorageBackend

logger = logging.getLogger('client')
logger.setLevel(logging.DEBUG)

class ChromeDriver(BaseDriver):
    
    def __new__(cls,
                *args,
                option_arguments: List[str] = None,
                services_kwargs: dict[str, Any] = None,
                keep_alive: bool = True,
                extensions: List[str | Path] = None,
                capabilities: Dict = None,
                quit_on_failure: bool = False,
                timeout: int = 20,
                remote: bool = False,
                **kwargs,) -> BaseDriver:
        
        browser = 'chrome'
        if remote:
            browser = 'chrome_remote'

        if 'stealth' in kwargs:
            kwargs['profile'] = kwargs.get('profile', 'Profile 1')
            kwargs['localstorage'] = kwargs.get('localstorage', LOCALSTORAGE)
        profile = kwargs.get('profile', None)
        localstorage = kwargs.get('localstorage', None)
        
        instance = super().__new__(
            cls,
            *args,
            browser=browser,
            option_arguments=option_arguments,
            services_kwargs=services_kwargs,
            keep_alive=keep_alive,
            extensions=extensions,
            capabilities=capabilities,
            **kwargs,
        )
        
        super().__init__(
            instance,
            quit_on_failure=quit_on_failure,
            timeout=timeout,
            profile=profile,
            localstorage=localstorage,
        )
        return instance

    def execute_cdp_cmd(self: WebDriver, cmd: str, cmd_args: dict):
        """
        Method copied from selenium.webdriver.chromium.webdriver.ChromiumDriver
        This is necessary in case of remote=True at consctructor, since in this
        case self will be instance of 
        selenium.webdriver.remote.webdriver.WebDriver which does not have it
        implemented.
        
        Execute Chrome Devtools Protocol command and get returned result The
        command and command args should follow chrome devtools protocol
        domains/commands, refer to link
        https://chromedevtools.github.io/devtools-protocol/

        :Args:
         - cmd: A str, command name
         - cmd_args: A dict, command args. empty dict {} if there is no
         command args
        :Usage:
            ::

                driver.execute_cdp_cmd(
                    'Network.getResponseBody', {'requestId': requestId}
                )
        :Returns:
            A dict, empty dict {} if there is no result to return.
            For example to getResponseBody:
            {'base64Encoded': False, 'body': 'response body string'}
        """
        return self.execute("executeCdpCommand", {"cmd": cmd, "params": cmd_args})["value"]

    def change_download_dir(self, path: str | Path):
        params = {
            "behavior": "allow",
            "downloadPath": str(path)
        }
        self.execute_cdp_cmd("Page.setDownloadBehavior", params)



import json
import platform
from os import name
from pathlib import Path
from copy import deepcopy
from itertools import cycle
from datetime import datetime
from random import choice, shuffle
from typing import Any, List, Optional, Dict

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def copy_list(original_list):
    """
    Returns a deep copy of a list containing dictionaries or strings.

    Args:
    original_list (list): The list to copy.

    Returns:
    list: A deep copy of the original list.
    """
    return deepcopy(original_list)


def delete_from_list(list_of_dicts, dict_item):
    """
    Removes all occurrences of dict_item from list_of_dicts.

    Args:
    list_of_dicts (list): List of dictionaries or strings.
    dict_item (dict or str): Item to remove from the list.

    Returns:
    list: List with dict_item removed.
    """
    return [item for item in list_of_dicts if item != dict_item]


def datetime_to_str(when):

    return when.isoformat()


# https://stackoverflow.com/questions/27522626/hash-function-in-python-3-3-returns-different-results-between-sessions
def hash_string(text: str) -> int:
    """
    Custom hash function for strings.

    Args:
    text (str): Input string.

    Returns:
    int: Hashed integer value.
    """
    hash_value = 0
    for ch in text:
        hash_value = (hash_value * 281 ^ ord(ch) * 997) & 0xFFFFFFFF
    return hash_value


class BaseData:
    def __init__(self):
        self.has_initialized = True
        self.data = self.get_data()
        self.set_data(self.data)

    def get_data(self) -> List[Any]:
        """Abstract method to get data. Should be overridden in subclass."""
        raise NotImplementedError

    @property
    def has_items(self) -> bool:
        return bool(self.data)

    def set_data(self, data: List[Any]) -> None:
        """Initialize or reset data and its cycled version."""
        self.data = data
        shuffled_data = list(data)
        shuffle(shuffled_data)
        self.cycled_data = cycle(shuffled_data)

    def get_random_cycled(self) -> Optional[Any]:
        """Get a random item from cycled data."""
        if self.has_items:
            return next(self.cycled_data)
        return None

    def get_random(self) -> Any:
        """Get a random item from data."""
        return choice(self.data)

    def remove_data(self, item: Any) -> None:
        """Remove an item from data and reset the cycled data."""
        self.data = [d for d in self.data if d != item]
        self.set_data(self.data)

    def get_hashed(self, value: Optional[Any]) -> Any:
        """Get a data item based on a hashed value."""
        if value is None:
            value = "_"
        hashed_value = hash(value)
        return self.data[hashed_value % len(self.data)]

    def get_n(self, n: int) -> List[Any]:
        """Get a list of n random cycled items."""
        return [self.get_random_cycled() for _ in range(n)]

    def get_hundred(self) -> List[Any]:
        """Get a list of 100 random cycled items."""
        return self.get_n(100)


def get_correct_agent(windows: str, mac: str, linux: str) -> str:
    """Selects the correct user agent based on the operating system."""
    if name == 'nt':
        return windows
    elif platform == "darwin":
        return mac
    return linux

def generate_user_agents(start_version: int,
                         end_version: int) -> Dict[str, str]:
    """Generates a dictionary of user agents for versions in the specified range."""
    return {
        str(version): get_correct_agent(
            f"Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version}.0.0.0 Safari/537.36",
            f"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version}.0.0.0 Safari/537.36",
            f"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version}.0.0.0 Safari/537.36",
        )
        for version in range(start_version, end_version + 1)
    }


class UserAgent(BaseData):
    
    REAL = "REAL"
    RANDOM = "RANDOM"
    HASHED = "HASHED"

    GOOGLE_BOT = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
    
    # Use dynamically generated user agents
    USER_AGENTS = generate_user_agents(98, 127)
    FREQUENCIES = {str(version): 1 for version in range(98, 128)}
    FREQUENCIES.update(
        {
            "127": 42,
            "126": 22, "125": 18, "124": 31, "123": 26,
            "122": 22, "121": 18, "120": 31, "119": 26,
            "118": 42, "117": 10, "116": 10, "115": 5,
            "114": 15, "113": 28, "112": 11, "111": 5,
            "110": 28, "109": 42, "108": 22, "107": 12,
            "106": 27, "105": 32, "104": 2, "103": 2,
            "102": 5, "101": 8, "100": 12,
            "99": 3, "98": 2
        }
    )

    def get_data(self) -> List[str]:
        """Returns a list of user agents based on predefined versions and their frequencies."""
        return [
            agent
            for version, count in self.FREQUENCIES.items()
            for agent in [self.USER_AGENTS[version]] * count
        ]


class WindowSize(BaseData):

    RANDOM = "RANDOM"
    HASHED = "HASHED"

    window_size_1920_1080 = [1920, 1080, ]
    window_size_1366_768 = [1366, 768, ]
    window_size_1536_864 = [1536, 864, ]
    window_size_1280_720 = [1280, 720, ]
    window_size_1440_900 = [1440, 900, ]
    window_size_1600_900 = [1600, 900, ]

    def get_data(self):

        # Windows
        N_1920_1080 = 35
        N_1366_768 = 26
        N_1536_864 = 16
        N_1280_720 = 9
        N_1440_900 = 9
        N_1600_900 = 5
        _1920_1080 = [self.window_size_1920_1080] * N_1920_1080
        _1366_768 = [self.window_size_1366_768] * N_1366_768
        _1536_864 = [self.window_size_1536_864] * N_1536_864
        _1280_720 = [self.window_size_1280_720] * N_1280_720
        _1440_900 = [self.window_size_1440_900] * N_1440_900
        _1600_900 = [self.window_size_1600_900] * N_1600_900

        result = _1920_1080 + _1366_768 + _1536_864 + _1280_720 + _1440_900 + _1600_900
        return result
    
    def to_string(self, window_size):
        width, height = window_size
        return f'{width},{height}'



class JSONStorageBackend:
    def __init__(self, base_path: Path = None):
        self.base_path = base_path or Path(".")
        self.json_path = self.base_path / "profiles.json"
        self.json_data = {}
        self.refresh()

    def refresh(self):
        if not self.json_path.is_file():
            self.commit_to_disk()

        with self.json_path.open("r") as json_file:
            self.json_data = json.load(json_file)

    def commit_to_disk(self):
        with self.json_path.open("w") as json_file:
            json.dump(self.json_data, json_file, indent=4)

    def get_item(self, key: str, default=None) -> Any:
        return self.json_data.get(key, default)

    def items(self):
        return self.json_data

    def set_item(self, key: str, value: Any) -> None:
        if "created_at" not in value:
            value["created_at"] = datetime_to_str(datetime.now())

        value["updated_at"] = datetime_to_str(datetime.now())

        self.json_data[key] = {"profile_id": key, **value}
        self.commit_to_disk()

    def remove_item(self, key: str) -> None:
        if key in self.json_data:
            self.json_data.pop(key)
            self.commit_to_disk()

    def clear(self) -> None:
        if self.json_path.is_file():
            self.json_path.unlink()
        self.json_data = {}
        self.commit_to_disk()


class ProfileStorageBackend(JSONStorageBackend):
    def __init__(self, base_path: Path = None):
        super().__init__(base_path)

    def get_profile(self, profile_name: str, default=None) -> Any:
        return self.get_item(profile_name, default)

    def set_profile(self, profile_name: str, profile_data: Any) -> None:
        profile_data["updated_at"] = datetime_to_str(datetime.now())
        self.set_item(profile_name, profile_data)

    def remove_profile(self, profile_name: str) -> None:
        self.remove_item(profile_name)

    def clear(self) -> None:
        super().clear()




from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import EdgeChromiumDriverManager

from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait

from selenium.webdriver.remote.webdriver import WebDriver as SeleniumWebDriver
from selenium.webdriver.remote.webelement import WebElement

from selenium.webdriver import Firefox
from selenium.webdriver import Chrome
from selenium.webdriver import Safari
from selenium.webdriver import Edge

from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.safari.service import Service as SafariService
from selenium.webdriver.edge.service import Service as EdgeService

from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.safari.options import Options as SafariOptions
from selenium.webdriver.edge.options import Options as EdgeOptions

from weberist.generic.constants import OPERATING_SYSTEM

if 'win' in OPERATING_SYSTEM:
    from selenium.webdriver import FirefoxOptions  # noqa F811

__all__ = (
    "Firefox",
    "Chrome",
    "Safari",
    "Edge",
    "FirefoxOptions",
    "ChromeOptions",
    "SafariOptions",
    "EdgeOptions",
    "FirefoxService",
    "ChromeService",
    "SafariService",
    "EdgeService",
    "WebElement",
    "expected_conditions",
    "WebDriverWait",
    "ChromeDriverManager",
    "GeckoDriverManager",
    "EdgeChromiumDriverManager",
    "SeleniumWebDriver",
)



import logging
import shutil
import binascii
from pathlib import Path

logger = logging.getLogger("weberist.base.stealth")


def hex_dump_to_bytes(hex_dump: str) -> bytes:
    # Convert a hex dump to bytes
    hex_str = "".join(hex_dump.split())
    return binascii.unhexlify(hex_str)


def remove_cdc(chromedriver_path: str):

    chromedriver_path = Path(chromedriver_path)

    # Create a backup of the original ChromeDriver
    backup_path = chromedriver_path.with_suffix(".bak")
    modified_path = chromedriver_path.with_name(chromedriver_path.name + ".modified")
    shutil.copy(chromedriver_path, backup_path)

    with chromedriver_path.open("rb") as binary_file:
        binary_content = binary_file.read()

    # Create a hex dump of the binary content
    hex_dump = binascii.hexlify(binary_content).decode("ascii")
    logger.info("Found %d occurrences of '$cdc_' and %d occurrences of 'webdriver'", hex_dump.count('246364635f'), hex_dump.count('777562647276657265'))

    # Replace occurrences of '$cdc_' with 'xymu' in the hex dump
    hex_dump = hex_dump.replace("246364635f", "7879646d75")
    # Replace occurrences of 'webdriver' with 'xyzabc'
    hex_dump = hex_dump.replace("777562647276657265", "78797a616263")

    # Convert the modified hex dump back to binary
    modified_binary_content = hex_dump_to_bytes(hex_dump)
    try:
        # Convert the modified hex dump back to binary
        modified_binary_content = hex_dump_to_bytes(hex_dump)

        with modified_path.open("wb") as binary_file:
            binary_file.write(modified_binary_content)

        # Replace the original file with the modified file
        shutil.move(modified_path, chromedriver_path)

        chromedriver_path.chmod(0o755)
        logger.info("Modified ChromeDriver at %s", chromedriver_path)

        with chromedriver_path.open("rb") as binary_file:
            binary_content = binary_file.read()
        hex_dump = binascii.hexlify(binary_content).decode("ascii")
        logger.info("Found %d occurrences of '$cdc_' and %d occurrences of 'webdriver'", hex_dump.count('246364635f'), hex_dump.count('777562647276657265'))

    except Exception as e:
        logger.error("An error occurred while modifying ChromeDriver: %s", e)
        # Clean up the modified file if something goes wrong
        if modified_path.exists():
            modified_path.unlink()



import re
import logging
import asyncio
import traceback
from typing import List
from urllib.parse import urlparse

logger = logging.getLogger('weberist.generic.utils')

async def cancel_task(task: asyncio.Task):
    if not task.done():
        task_name = task.get_name()
        try:
            task.cancel()
            logger.debug("Cancelling task %s", task_name)
            await asyncio.sleep(0.1)
        except asyncio.CancelledError as err:
            logger.debug("Cancelling task %s failed: ", err)

async def gather_and_handle(tasks: List[asyncio.Task],
                            results: List = None,
                            raise_error: bool = True):
    done, pending = await asyncio.wait(
        tasks, return_when=asyncio.FIRST_EXCEPTION
    )
    task_name = ''
    if results is None:
        results = []
    if done:
        for task in done:
            err = task.exception()
            if err:
                task_name = task.get_name()
                logger.error(
                    'Task %s failed. ERROR: %s. Traceback: %s',
                    task_name,
                    err,
                    traceback.format_exc()
                )
                await cancel_task(task)
                if raise_error:
                    for task_ in tasks:
                        await cancel_task(task_)
                    raise err
                results.append(None)
                continue
            results.append(task.result())
    if pending:
        return await gather_and_handle(pending, results, raise_error)
    return results


def run_async(coro, *args, **kwargs):
    try:
        loop = asyncio.get_running_loop()
        return loop.run_until_complete(coro(*args, **kwargs))
    except RuntimeError:
        try:
            return asyncio.run(coro(*args, **kwargs))
        except RuntimeError:
            try:
                import nest_asyncio  # pylint --disable=import-outside-toplevel
                nest_asyncio.apply()
                return asyncio.run(coro(*args, **kwargs))
            except (ImportError, ModuleNotFoundError) as err:
                logger.error(err)
                logger.warning("Install nest_asyncio to run nested async.")
                raise err

def extract_base_url(text: str) -> str:
    url_pattern = r'(https?://[^\s]+)'
    urls = re.findall(url_pattern, text)

    if not urls:
        return None

    parsed_url = urlparse(urls[0])
    base_url = f'{parsed_url.scheme}://{parsed_url.netloc}'
    
    return base_url


from urllib3.exceptions import MaxRetryError
from requests.adapters import ReadTimeout
from requests.adapters import ConnectionError as RequestConnectionError
from selenium.webdriver.remote.errorhandler import NoSuchWindowException
from selenium.webdriver.remote.errorhandler import NoSuchElementException
from selenium.webdriver.remote.errorhandler import InvalidSessionIdException
from selenium.webdriver.remote.errorhandler import StaleElementReferenceException  # noqa E501
from selenium.webdriver.remote.errorhandler import ElementNotInteractableException  # noqa E501
from selenium.webdriver.remote.errorhandler import ElementClickInterceptedException  # noqa E501
from selenium.common.exceptions import WebDriverException

# exceptions for quit_on_failure
EXCEPTIONS = (
    ReadTimeout,
    WebDriverException,
    NoSuchWindowException,
    RequestConnectionError,
    NoSuchElementException,
    NoSuchElementException,
    InvalidSessionIdException,
    StaleElementReferenceException,
    ElementNotInteractableException,
    ElementClickInterceptedException,
    Exception,
)

__all__ = (
    "MaxRetryError",
    "ReadTimeout",
    "RequestConnectionError",
    "NoSuchWindowException",
    "NoSuchElementException",
    "StaleElementReferenceException",
    "ElementNotInteractableException",
    "ElementClickInterceptedException",
    "WebDriverException",
)



import os
import json
import time
import signal
import asyncio
import asyncio
import logging
import tarfile
import threading
import tarfile
import threading
import subprocess
from io import BytesIO
from io import BytesIO
from shutil import copyfile
from pathlib import Path
from typing import Callable, List, Dict

import docker

from weberist import ChromeDriver
from weberist.base.config import (
    DATA_DIR,
    DOCKER_DIR,
    DOCKER_FILE_BROWSER,
    LOCALSTORAGE,
    BROWSER_IMAGE,
    LOCALSTORAGE,
    BROWSER_IMAGE,
    CHROME_IMAGE,
    DOCKER_NETWORK,
    CHROME_VERSIONS,
    FIREFOX_VERSIONS,
    DOCKER_COMPOSE,
    CONTAINER_SELENOID,
    CONTAINER_SELENOID_UI,
    BROWSER_DICT,
)
from weberist.generic.types import TypeBrowser
from weberist.generic.utils import run_async, gather_and_handle


SELENOID_STARTED_CUE = "[INIT] [Listening on :4444]"
is_selenoid_up = False

logger = logging.getLogger('standard')
client_logger = logging.getLogger('client')
logger.setLevel(logging.DEBUG)
client_logger.setLevel(logging.DEBUG)


def create_network(name: str = None, client: docker.DockerClient = None):

    if name is None:
        name = DOCKER_NETWORK
    if client is None:
        client = docker.from_env()
    for network in client.networks.list():
        if name == network.name:
            client_logger.warning("Network '%s' already exists.", name)
            return network
    return client.networks.create(name)

def create_dockerfile(name: str = None,
                      browser: str = None,
                      version: str = None,
                      target_path: str | Path = None):
    
    target_path = target_path or DATA_DIR
    if not isinstance(target_path, Path):
        target_path = Path(target_path)

    with open(DOCKER_FILE_BROWSER, 'r', encoding='utf-8') as dockerfile_chrome:
        localstorage = target_path / 'localstorage'
        localstorage.mkdir(parents=True, exist_ok=True)
        entrypoint = target_path / f'{browser}-entrypoint.sh'
        copyfile(DOCKER_DIR / f'{browser}-entrypoint.sh', entrypoint)
        dockerfile_content = dockerfile_chrome.read().format(
            version=version,
            browser=browser,
            localstorage='./localstorage',
            entrypoint=f'./{browser}-entrypoint.sh'
        )

        name = name or 'Dockerfile'
        with open(target_path / name, 'w', encoding='utf-8') as dockerfile:
            dockerfile.write(dockerfile_content)
    
def create_chrome_dockerfile(name: str = None,
                             chrome_version: str = None,
                             target_path: str | Path = None):

    chrome_version = chrome_version or CHROME_VERSIONS[-1]
    create_dockerfile(name, 'chrome', chrome_version, target_path)

def create_firefox_dockerfile(name: str = None,
                              firefox_version: str = None,
                              target_path: str | Path = None):

    firefox_version = firefox_version or FIREFOX_VERSIONS[-1]
    create_dockerfile(name, 'firefox', firefox_version, target_path)
                        
def create_browser_image(browser: str,
                         version: int,
                         client: docker.DockerClient = None,
                         target_path: str | Path = None):

    image, log = None, None
    create = True
    client = client or docker.from_env()
    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    tag = BROWSER_IMAGE.format(
        browser=browser, version=version
    )
    # tag = f"{tag}:latest"
    for image_ in client.images.list():
        if image_ and len(image_.tags) > 0:
            if tag == image_.tags[0]:
                create = False
                image = image_
                break
    if create:
        client_logger.info("Creating image '%s'", tag)
        
        name = f"Dockerfile-{browser}-{version}"
        create_dockerfile(
            name=name,
            browser=browser,
            version=version,
            target_path=target_path
        )
        dockerfile_obj = BytesIO()
        with tarfile.open(fileobj=dockerfile_obj, mode='w') as tar:
            for file_path in target_path.rglob('*'):
                arcname = file_path.relative_to(target_path)
                tar.add(file_path, arcname=arcname)
        dockerfile_obj.seek(0)
        
        image, log = client.images.build(
            fileobj=dockerfile_obj,
            dockerfile=name,
            custom_context=True,
            tag=tag,
            rm=True,
            nocache=True,
            quiet=False,
        )
        
        client_logger.info("Image '%s' crated!", tag)
    
    return image, log
                        
async def create_browser_image_async(browser: str,
                                     version: int,
                                     client: docker.DockerClient = None,
                                     target_path: str | Path = None):
    return await asyncio.to_thread(
        create_browser_image,
        browser,
        version,
        client,
        target_path,
    )

def create_browsers_images(browsers: Dict[str, TypeBrowser],
                           client: docker.DockerClient = None,
                           target_path: str | Path = None):

    data = {
        "images": [],
        "logs": []
    }

    client = client or docker.from_env()
    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    for browser, info in browsers.items():
        for version in info['versions']:
            image, log = create_browser_image(
                browser, version, client, target_path
            )
            data['images'].append(image)
            data['logs'].append(log)
    return data

async def create_browsers_images_async(browsers: Dict[str, TypeBrowser],
                                       client: docker.DockerClient = None,
                                       target_path: str | Path = None,
                                       n_batches: int = 4):

    data = {
        "images": [],
        "logs": []
    }

    client = client or docker.from_env()
    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    
    tasks = []
    results = []
    counter = 0
    for browser, info in browsers.items():
        for version in info['versions']:
            tasks.append(
                asyncio.create_task(
                    create_browser_image_async(
                        browser, version, client, target_path
                    ),
                    name=f"{browser}-{version}"
                )
            )
            counter += 1
            if counter == n_batches:
                result = await gather_and_handle(tasks, raise_error=False)
                results.extend(result)
                tasks = []
    if len(tasks) > 0:
        result = await gather_and_handle(tasks, raise_error=False)
        results.extend(result)
    for result in results:
        if result is not None:
            image, log = result
            data['images'].append(image)
            data['logs'].append(log)
    return data

def create_browser_config_dict(browser: str,
                               versions: List[int],
                               default: int = 0,
                               **kwargs):
    browser_config = {}
    images = [f"{browser}_{version}.0" for version in versions]
    browser_config["default"] = images[default]
    browser_config["versions"] = {}
    for image in images:
        browser_config["versions"][image] = BROWSER_DICT
        browser_config["versions"][image]["image"] = f"weberist-{image}"
        for key, value in kwargs.items():
            browser_config["versions"][image][key] = value
    return browser_config

def create_browsers_json(browsers: Dict[str, TypeBrowser],
                         target_path: str | Path = None):
    browsers_json = {}
    for browser, info in browsers.items():
        kwargs = {}
        if 'kwargs' in info:
            kwargs = info['kwargs']
        browsers_json[browser] = create_browser_config_dict(
            browser,
            info['versions'],
            info['default'],
            **kwargs
        )
    target_path = target_path or DATA_DIR
    if not isinstance(target_path, Path):
        target_path = Path(target_path)

    path = target_path / 'browsers.json'
    with open(path, 'w', encoding='utf-8') as json_file:
        json.dump(browsers_json, json_file, indent=4, ensure_ascii=False)

def create_selenoid_compose(browsers: Dict[str, TypeBrowser],
                            name: str = None,
                            network_name: str = None,
                            client: docker.DockerClient = None,
                            target_path: str | Path = None,):

    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    target_path.mkdir(parents=True, exist_ok=True)
    target = target_path / 'target'
    video = target_path / 'video'
    logs = target_path / 'logs'
    target.mkdir(parents=True, exist_ok=True)
    video.mkdir(parents=True, exist_ok=True)
    logs.mkdir(parents=True, exist_ok=True)
    
    create_browsers_json(browsers, target_path)
    
    # data = create_browsers_images(browsers, client, target_path)
    data = run_async(
        create_browsers_images_async, browsers, client, target_path
    )
    name = name or DOCKER_COMPOSE
    network_name = network_name or DOCKER_NETWORK

    template_compose = DOCKER_DIR / "docker-compose-selenoid.yml"
    with open(template_compose, 'r', encoding='utf-8') as docker_compose_file:
        content = docker_compose_file.read().format(network=network_name)
        target_name = target_path / name
        with open(target_name, 'w', encoding='utf-8') as compose_file:
            compose_file.write(content)
    
    return data
                                   
def create_selenoid_chrome_compose(name: str = None,
                                   dockerfile_name: str = None,
                                   network_name: str = None,
                                   chrome_version: str = None,
                                   target_path: str | Path = None):

    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    target_path.mkdir(parents=True, exist_ok=True)
    target = target_path / 'target'
    video = target_path / 'video'
    logs = target_path / 'logs'
    target.mkdir(parents=True, exist_ok=True)
    video.mkdir(parents=True, exist_ok=True)
    logs.mkdir(parents=True, exist_ok=True)

    name = name or DOCKER_COMPOSE
    network_name = network_name or DOCKER_NETWORK

    create_chrome_dockerfile(dockerfile_name, chrome_version, target_path)
    create_browsers_json(chrome_version, target_path)

    template_compose = DOCKER_DIR / "docker-compose-selenoid.yml"
    with open(template_compose, 'r', encoding='utf-8') as docker_compose_file:

        dockerfile_content = docker_compose_file.read().format(
            network=network_name,
            # NOTE: conflict if already created:
            container_selenoid=CONTAINER_SELENOID,
            container_selenoid_ui=CONTAINER_SELENOID_UI,
        )
        target_name = target_path / name
        with open(target_name, 'w', encoding='utf-8') as target:
            target.write(dockerfile_content)
        
                        
def create_chrome_image(chrome_version: str = CHROME_VERSIONS[-1],
                        client: docker.DockerClient = None,
                        target_path: str | Path = None):

    client = client or docker.from_env()
    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    image = None
    log = [{}]
    create = True
    name = f"{CHROME_IMAGE.format(version=chrome_version)}:latest"
    for image_ in client.images.list():
        if image_ and len(image_.tags) > 0:
            if name == image_.tags[0]:
                create = False
                image = image_
                break
    if create:
        client_logger.info("Creating chrome image '%s'", name)
        image, log = client.images.build(
            path=str(target_path.absolute()),
            tag=name,
            rm=True,
            nocache=True,
            quiet=False,
        )
        client_logger.info("Chrome image '%s' crated!", name)
    return image, log

def setup_selenoid(browsers: Dict[str, TypeBrowser],
                   network_name: str = None,
                   target_path: str | Path = None,
                   client: docker.DockerClient = None):

    client = client or docker.from_env()
    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)

    network = create_network(name=network_name, client=client)
    data = create_selenoid_compose(browsers, network_name, target_path)
    
    data['network'] = network
    
    return data

def setup_selenoid_chrome(dockerfile_name: str = None,
                          dockercompose_name: str = None,
                          network_name: str = None,
                          chrome_version: str = CHROME_VERSIONS[-1],
                          target_path: str | Path = None,
                          client: docker.DockerClient = None):

    client = client or docker.from_env()
    target_path = target_path or DATA_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)

    network = create_network(name=network_name, client=client)

    create_selenoid_chrome_compose(
        dockercompose_name,
        dockerfile_name,
        network_name,
        chrome_version,
        target_path,
    )

    image, log = create_chrome_image(chrome_version, client)

    return image, network, log

def run_docker_compose(path: str = None, build: bool = False):
    # Start the Docker Compose process
    command = ["docker", "compose", "up", "--build"]
    if path is not None:
        command = command[:2] + ["-f", str(path)] + command[2:]
    if not build:
        command.pop()
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,  # Use text mode for easier handling of strings
        bufsize=1,  # Line buffering
        universal_newlines=True  # Ensure newlines are handled properly
    )

    # Define a function to print output in real-time
    def print_output(pipe, stream):
        for line in iter(pipe.readline, ''):
            if stream == 'stdout':
                client_logger.info("stdout: %s", line)
            else:
                client_logger.info("stderr: %s", line)
            if SELENOID_STARTED_CUE in line:
                global is_selenoid_up
                is_selenoid_up = True
        pipe.close()

    # Start threads to handle stdout and stderr
    stdout_thread = threading.Thread(target=print_output, args=(process.stdout, 'stdout'))
    stderr_thread = threading.Thread(target=print_output, args=(process.stderr, 'stderr'))

    stdout_thread.start()
    stderr_thread.start()

    # Handle process termination and cleanup
    def cleanup():
        if process.poll() is None:  # Check if process is still running
            os.kill(process.pid, signal.SIGTERM)  # Send termination signal
            process.wait()  # Wait for the process to terminate

    return process, stdout_thread, stderr_thread, cleanup

def stop_docker_compose(process, path: str = None):
    # Stop the Docker Compose process
    process.terminate()
    try:
        process.wait(timeout=10)
    except subprocess.TimeoutExpired:
        process.kill()

    # Optionally, you can shut down the containers cleanly
    command = ["docker", "compose", "down"]
    if path is not None:
        command = command[:2] + ["-f", str(path)] + command[2:]
    subprocess.run(command, check=True)

def wait_selenoid(quiet=False):

    global is_selenoid_up

    while not is_selenoid_up:
        time.sleep(1)
        if not quiet:
            client_logger.debug("🕒 Waiting selenoid...")


def run_selenoid_driver_task(driver_task: Callable,
                             *args,
                             dockercompose_name: str = None,
                             network_name: str = None,
                             target_path: str | Path = None,
                             chrome_kwargs: dict = None,
                             **kwargs):

    global is_selenoid_up
    is_selenoid_up = False

    client = docker.from_env()
    containers = client.containers.list()
    for container in containers:
        if 'selenoid' in container.name:
            is_selenoid_up = True
            break

    dockercompose_name = dockercompose_name or DOCKER_COMPOSE
    target_path = target_path or DATA_DIR  # DOCKER_DIR
    if isinstance(target_path, str):
        target_path = Path(target_path)
    browsers = {
        'chrome': TypeBrowser(versions=[CHROME_VERSIONS[-1]], default=-1)
    }
    setup_selenoid(
        browsers,
        network_name,
        target_path,
        client
    )

    path = None
    process = None
    if not is_selenoid_up:
        client_logger.debug("Selenoid is not up")

        build = True
        path = Path(target_path) / dockercompose_name
        for container in client.containers.list(all=True):
            if CONTAINER_SELENOID == container.name:
                build = False
                break
        (
            process, stdout_thread, stderr_thread, clean_up
        ) = run_docker_compose(path, build)
        wait_selenoid()
    else:
        client_logger.debug("Selenoid is already up")

    try:
        if chrome_kwargs is None:
            chrome_kwargs = {}
        if 'localstorage' not in chrome_kwargs:
            chrome_kwargs['localstorage'] = LOCALSTORAGE
        driver = ChromeDriver(remote=True, **chrome_kwargs)
        result = driver_task(driver, *args, **kwargs)
        return result
    finally:
        if process:
            stop_docker_compose(process, path)
            clean_up()
            stdout_thread.join()
            stderr_thread.join()



"""Configurations for development and maintenance"""

from pathlib import Path

ROOT_DIR = Path(__file__).resolve().parent.parent
DATA_DIR = ROOT_DIR / 'data'
DOCKER_DIR = ROOT_DIR / 'docker'
LOCALSTORAGE = DATA_DIR / 'localstorage'
DOCKER_CHROME_LOCALSTORAGE = ROOT_DIR / 'localstorage'
CHROME_EXTENSIONS = DATA_DIR / 'extensions/chrome'
DOCKER_FILE_BROWSER = DOCKER_DIR / 'Dockerfile'
DOCKER_FILE_CHROME = DOCKER_DIR / 'Dockerfile-chrome'
BROWSER_IMAGE = 'weberist-{browser}_{version}.0'
CHROME_IMAGE = 'weberist-chrome_{version}.0'
DOCKER_COMPOSE = 'docker-compose.yml'
DOCKER_NETWORK = 'weberist'
CONTAINER_SELENOID = 'weberist-selenoid'
CONTAINER_SELENOID_UI = 'weberist-selenoid-ui'
DEFAULT_PROFILE = 'Profile 1'
CHROME_VERSIONS = tuple(str(i) for i in range(48, 128))
FIREFOX_VERSIONS = tuple(str(i) for i in range(4, 125))

BROWSER_DICT = {
    "image": None,
    "port": "4444",
    "tmpfs": {
        "/tmp": "size=512m",
        "/var": "size=128m",
    },
    "path": "/"
}

LOG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "client": {
            "format": (
                "%(levelname)s (%(filename)s at line %(lineno)d): %(message)s"
            )
        },
        "standard": {
            "format": (
                "%(levelname)s (%(funcName)s): %(message)s"
                "\n\t├─file: %(pathname)s"
                "\n\t╰─line: %(lineno)d"
            )
        },
        "debug": {
            "format": (
                "%(asctime)s %(levelname)s (at %(funcName)s "
                "in line %(lineno)d):"
                "\n\t├─file: %(pathname)s"
                "\n\t├─task name: %(taskName)s"
                "\n\t╰─message: %(message)s\n"
            ),
            "datefmt": "%y-%m-%d %H:%M:%S"
        }
    },
    "handlers": {
        "client": {
            "class": "logging.StreamHandler",
            "formatter": "client",
        },
        "standard": {
            "class": "logging.StreamHandler",
            "formatter": "standard",
        },
        "debug": {
            "class": "logging.StreamHandler",
            "formatter": "debug",
        }
    },
    "loggers": {
        "": {"handlers": ["client"], "level": "DEBUG"},
        "standard": {
            "handlers": ["standard"],
            "level": "DEBUG",
            "propagate": False,
        },
        "debugger": {
            "handlers": ["debug"],
            "level": "DEBUG",
            "propagate": False,
        }
    }
}